/*
 * Sony TV tuner driver
 *
 * Copyright (C) 2005-2006 Micronas USA Inc.
 * Copyright (C) 2009-2010 Sensoray Co., Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (Version 2) as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/videodev2.h>
#include <media/tuner.h>
#include <media/v4l2-common.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-device.h>
#include <media/v4l2-i2c-drv.h>

MODULE_DESCRIPTION("Sony TV Tuner driver");
MODULE_LICENSE("GPL v2");

static int debug;
module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "debug level 0=off(default) 1=on\n");

/* #define MPX_DEBUG */

/* AS(IF/MPX) pin:      LOW      HIGH/OPEN
 * IF/MPX address:   0x42/0x40   0x43/0x44
 */
#define IF_I2C_ADDR	0x43
#define MPX_I2C_ADDR	0x44

static v4l2_std_id force_band;
static char force_band_str[] = "-";
module_param_string(force_band, force_band_str, sizeof(force_band_str), 0644);

/* Store tuner info in the same format as tuner.c, so maybe we can put the
 * Sony tuner support in there. */
struct sony_tunertype {
	const char *name;
	unsigned char Vendor; /* unused here */
	unsigned char Type; /* unused here */

	unsigned short thresh1; /*  band switch VHF_LO <=> VHF_HI */
	unsigned short thresh2; /*  band switch VHF_HI <=> UHF */
	unsigned char VHF_L;
	unsigned char VHF_H;
	unsigned char UHF;
	unsigned char config;
	unsigned short IFPCoff;
};

/* This array is indexed by (tuner_type - TUNER_SONY_BTF_PG472Z) */
static const struct sony_tunertype sony_tuners[] = {
	{ "Sony PAL+SECAM (BTF-PG472Z)", 0, 0,
	  16*144.25, 16*427.25, 0x01, 0x02, 0x04, 0xc6, 623},
	{ "Sony NTSC_JP (BTF-PK467Z)", 0, 0,
	  16*220.25, 16*467.25, 0x01, 0x02, 0x04, 0xc6, 940},
	{ "Sony NTSC (BTF-PB463Z)", 0, 0,
	  16*130.25, 16*364.25, 0x01, 0x02, 0x04, 0xc6, 732},
};

struct sony_tuner {
	struct v4l2_subdev sd;
	int type;
	v4l2_std_id std;
	unsigned int freq;
	struct v4l2_subdev *mpx;
	u32 audmode;
};

static inline struct sony_tuner *to_state(struct v4l2_subdev *sd)
{
	return container_of(sd, struct sony_tuner, sd);
}

/* Basically the same as default_set_tv_freq() in tuner.c */
static int set_freq(struct i2c_client *client, int freq)
{
	struct sony_tuner *t = to_state(i2c_get_clientdata(client));
	const char *band_name;
	int n;
	int band_select;
	const struct sony_tunertype *tun;
	u8 buffer[4];

	tun = &sony_tuners[t->type - TUNER_SONY_BTF_PG472Z];
	if (freq < tun->thresh1) {
		band_name = "VHF_L";
		band_select = tun->VHF_L;
	} else if (freq < tun->thresh2) {
		band_name = "VHF_H";
		band_select = tun->VHF_H;
	} else {
		band_name = "UHF";
		band_select = tun->UHF;
	}
	v4l2_dbg(1, debug, client, "tuning to frequency %d.%04d (%s)\n",
			freq / 16, (freq % 16) * 625, band_name);
	n = freq + tun->IFPCoff;

	buffer[0] = n >> 8;
	buffer[1] = n & 0xff;
	buffer[2] = tun->config;
	buffer[3] = band_select;
	i2c_master_send(client, buffer, 4);

	return 0;
}


/*
 * IF configuration values for the BTF-PG472Z:
 *
 *	B/G: 0x94 0x70 0x49
 *	I:   0x14 0x70 0x4a
 *	D/K: 0x14 0x70 0x4b
 *	L:   0x04 0x70 0x4b
 *	L':  0x44 0x70 0x53
 *	M:   0x50 0x30 0x4c
 */

static int set_if(struct i2c_client *client)
{
	struct sony_tuner *t = i2c_get_clientdata(client);
	u8 buffer[4];
	struct i2c_msg msg;

	/* configure IF */
	buffer[0] = 0;
	if (t->std & V4L2_STD_PAL_BG) {
		buffer[1] = 0x94;
		buffer[2] = 0x70;
		buffer[3] = 0x49;
	} else if (t->std & V4L2_STD_PAL_I) {
		buffer[1] = 0x14;
		buffer[2] = 0x70;
		buffer[3] = 0x4a;
	} else if (t->std & V4L2_STD_PAL_DK) {
		buffer[1] = 0x14;
		buffer[2] = 0x70;
		buffer[3] = 0x4b;
	} else if (t->std & V4L2_STD_SECAM_L) {
		buffer[1] = 0x04;
		buffer[2] = 0x70;
		buffer[3] = 0x4b;
	}
	msg.addr = IF_I2C_ADDR;
	msg.flags = 0;
	msg.len = 4;
	msg.buf = buffer;
	i2c_transfer(client->adapter, &msg, 1);

	/* setup mpx */
	v4l2_subdev_call(t->mpx, core, s_std, t->std);

	return 0;
}

static int sony_tuner_s_type_addr(struct v4l2_subdev *sd,
				      struct tuner_setup *setup)
{
	struct sony_tuner *t = to_state(sd);

	if (!(setup->mode_mask & T_ANALOG_TV))
		return 0;

	if (t->type >= 0) {
		if (t->type != setup->type)
			v4l2_dbg(1, debug, sd, "type already set to %d, "
				"ignoring request for %d\n",
				t->type, setup->type);
		return 0;
	}
	switch (setup->type) {
	case TUNER_SONY_BTF_PG472Z:
		switch (force_band_str[0]) {
		case 'b':
		case 'B':
		case 'g':
		case 'G':
			v4l2_dbg(1, debug, sd, "forcing tuner to PAL-B/G bands\n");
			force_band = V4L2_STD_PAL_BG;
			break;
		case 'i':
		case 'I':
			v4l2_dbg(1, debug, sd, "forcing tuner to PAL-I band\n");
			force_band = V4L2_STD_PAL_I;
			break;
		case 'd':
		case 'D':
		case 'k':
		case 'K':
			v4l2_dbg(1, debug, sd, "forcing tuner to PAL-D/K bands\n");
			force_band = V4L2_STD_PAL_I;
			break;
		case 'l':
		case 'L':
			v4l2_dbg(1, debug, sd, "forcing tuner to SECAM-L band\n");
			force_band = V4L2_STD_SECAM_L;
			break;
		default:
			force_band = 0;
			break;
		}
		if (force_band)
			t->std = force_band;
		else
			t->std = V4L2_STD_PAL_BG;
		set_if(v4l2_get_subdevdata(sd));
		break;
	case TUNER_SONY_BTF_PK467Z:
		t->std = V4L2_STD_NTSC_M_JP;
		break;
	case TUNER_SONY_BTF_PB463Z:
		t->std = V4L2_STD_NTSC_M;
		break;
	default:
		v4l2_dbg(1, debug, sd,
			"tuner type %d is not supported by this module\n",
			setup->type);
		return -EINVAL;
	}
	t->type = setup->type;
	if (t->type >= 0)
		v4l2_dbg(1, debug, sd, "type set to %d (%s)\n",
			t->type,
			sony_tuners[t->type - TUNER_SONY_BTF_PG472Z].name);
	return 0;
}

static int sony_tuner_g_frequency(struct v4l2_subdev *sd,
				      struct v4l2_frequency *freq)
{
	struct sony_tuner *t = to_state(sd);

	if (freq->type != V4L2_TUNER_ANALOG_TV)
		return -EINVAL;

	freq->frequency = t->freq;
	return 0;
}

static int sony_tuner_s_frequency(struct v4l2_subdev *sd,
				      struct v4l2_frequency *freq)
{
	struct sony_tuner *t = to_state(sd);

	if (freq->type != V4L2_TUNER_ANALOG_TV)
		return -EINVAL;

	t->freq = freq->frequency;
	set_freq(v4l2_get_subdevdata(sd), t->freq);
	return 0;
}



static int sony_tuner_s_std(struct v4l2_subdev *sd, v4l2_std_id norm)
{
	struct sony_tuner *t = to_state(sd);
	v4l2_std_id old = t->std;

	switch (t->type) {
	case TUNER_SONY_BTF_PG472Z:
		if (force_band && (norm & force_band) != norm &&
				norm != V4L2_STD_PAL &&
				norm != V4L2_STD_SECAM) {
			v4l2_dbg(1, debug, sd, "ignoring requested TV standard in "
					"favor of force_band value\n");
			t->std = force_band;
		} else if (norm & V4L2_STD_PAL_BG) { /* default */
			t->std = V4L2_STD_PAL_BG;
		} else if (norm & V4L2_STD_PAL_I) {
			t->std = V4L2_STD_PAL_I;
		} else if (norm & V4L2_STD_PAL_DK) {
			t->std = V4L2_STD_PAL_DK;
		} else if (norm & V4L2_STD_SECAM_L) {
			t->std = V4L2_STD_SECAM_L;
		} else {
			v4l2_dbg(1, debug, sd, "TV standard not supported\n");
			return -EINVAL;
		}
		if (old != t->std)
			set_if(v4l2_get_subdevdata(sd));
		break;
	case TUNER_SONY_BTF_PK467Z:
		if (!(norm & V4L2_STD_NTSC_M_JP)) {
			v4l2_dbg(1, debug, sd, "TV standard not supported\n");
			return -EINVAL;
		}
		break;
	case TUNER_SONY_BTF_PB463Z:
		if (!(norm & V4L2_STD_NTSC_M)) {
			v4l2_dbg(1, debug, sd, "TV standard not supported\n");
			return -EINVAL;
		}
		break;
	}
	return 0;
}

static int sony_tuner_g_tuner(struct v4l2_subdev *sd, struct v4l2_tuner *vt)
{
	struct sony_tuner *t = to_state(sd);

	strlcpy(vt->name, "Television", sizeof(vt->name));
	vt->type = V4L2_TUNER_ANALOG_TV;
	/* Since we do not have a datasheet for this tuner we just
	 * put in artificial low and high frequency values. */
	vt->rangelow = 16;
	vt->rangehigh = 16 * 999.99;
	switch (t->type) {
	case TUNER_SONY_BTF_PG472Z:
		v4l2_subdev_call(t->mpx, tuner, g_tuner, vt);
		vt->capability |= V4L2_TUNER_CAP_NORM;
		break;
	case TUNER_SONY_BTF_PK467Z:
	case TUNER_SONY_BTF_PB463Z:
		vt->capability = V4L2_TUNER_CAP_STEREO;
		vt->rxsubchans = V4L2_TUNER_SUB_MONO | V4L2_TUNER_SUB_STEREO;
		break;
	}
	vt->audmode = t->audmode;
	return 0;
}

static int sony_tuner_s_tuner(struct v4l2_subdev *sd, struct v4l2_tuner *vt)
{
	struct sony_tuner *t = to_state(sd);

	if (vt->type != V4L2_TUNER_ANALOG_TV)
		return -EINVAL;

	switch (t->type) {
	case TUNER_SONY_BTF_PG472Z:
		if (vt->audmode != t->audmode) {
			t->audmode = vt->audmode;
			/* mpx_setup(v4l2_get_subdevdata(sd)); */
			v4l2_subdev_call(t->mpx, tuner, s_tuner, vt);
		}
		break;
	case TUNER_SONY_BTF_PK467Z:
	case TUNER_SONY_BTF_PB463Z:
		break;
	}
	return 0;
}

static int sony_tuner_log_status(struct v4l2_subdev *sd)
{
	struct sony_tuner *t = to_state(sd);

	v4l2_info(sd, "Standard: %s\n", t->std == V4L2_STD_NTSC ? "NTSC" :
					t->std == V4L2_STD_PAL ? "PAL" :
					t->std == V4L2_STD_SECAM ? "SECAM" :
					"unknown");
	v4l2_info(sd, "Frequency: %ud\n", t->freq);
	return 0;
}

/* --------------------------------------------------------------------------*/

static const struct v4l2_subdev_core_ops sony_tuner_core_ops = {
	.log_status = sony_tuner_log_status,
	.s_std = sony_tuner_s_std,
};

static const struct v4l2_subdev_tuner_ops sony_tuner_tuner_ops = {
	.s_frequency = sony_tuner_s_frequency,
	.g_frequency = sony_tuner_g_frequency,
	.s_tuner = sony_tuner_s_tuner,
	.g_tuner = sony_tuner_g_tuner,
	.s_type_addr = sony_tuner_s_type_addr,
};

static const struct v4l2_subdev_ops sony_tuner_ops = {
	.core = &sony_tuner_core_ops,
	.tuner = &sony_tuner_tuner_ops,
};

/* --------------------------------------------------------------------------*/

static int sony_tuner_probe(struct i2c_client *client,
				const struct i2c_device_id *id)
{
	struct sony_tuner *t;
	struct v4l2_subdev *sd;

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_SMBUS_I2C_BLOCK))
		return -ENODEV;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
	snprintf(client->name, sizeof(client->name) - 1, "sony_tuner");
#endif
	v4l_info(client, "chip found @ 0x%x (%s)\n",
		client->addr << 1, client->adapter->name);

	t = kzalloc(sizeof(struct sony_tuner), GFP_KERNEL);
	if (t == NULL)
		return -ENOMEM;

	sd = &t->sd;
	v4l2_i2c_subdev_init(sd, client, &sony_tuner_ops);

	/* Initialize sony_tuner */
	t->type = -1;
	t->freq = 0;
	t->audmode = V4L2_TUNER_MODE_STEREO;
	t->mpx = v4l2_i2c_new_subdev(sd->v4l2_dev, client->adapter,
		"sony_tuner_mpx", "sony_tuner_mpx", MPX_I2C_ADDR, NULL);

	return 0;
}

static int sony_tuner_remove(struct i2c_client *client)
{
	struct v4l2_subdev *sd = i2c_get_clientdata(client);

	v4l2_device_unregister_subdev(sd);
	kfree(to_state(sd));

	return 0;
}

/* ----------------------------------------------------------------------- */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
static const struct i2c_device_id sony_tuner_id[] = {
	{ "sony_tuner", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, sony_tuner_id);
#endif

static struct v4l2_i2c_driver_data v4l2_i2c_data = {
	.name = "sony_tuner",
	.probe = sony_tuner_probe,
	.remove = sony_tuner_remove,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
	.id_table = sony_tuner_id,
#endif
};
