/*
 * Copyright (C) 2005-2006 Micronas USA Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (Version 2) as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/videodev2.h>
#include <media/tvaudio.h>
#include <media/v4l2-common.h>
#include <media/v4l2-device.h>
#include <media/v4l2-i2c-drv.h>

MODULE_DESCRIPTION("UDA1342 I2C subdev driver");
MODULE_LICENSE("GPL v2");

static int write_reg(struct v4l2_subdev *sd, int reg, int value)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	/* UDA1342 wants MSB first, but SMBus sends LSB first */
	i2c_smbus_write_word_data(client, reg, swab16(value));
	return 0;
}

static int uda1342_s_audio_routing(struct v4l2_subdev *sd, u32 input,
				       u32 output, u32 config)
{
	switch (input) {
	case TVAUDIO_INPUT_TUNER:
		write_reg(sd, 0x00, 0x1441); /* select input 2 */
		break;
	case TVAUDIO_INPUT_EXTERN:
		write_reg(sd, 0x00, 0x1241); /* select input 1 */
		break;
	default:
		v4l2_err(sd, "input %d not supported\n", input);
		return -EINVAL;
	}
	return 0;
}

static int uda1342_log_status(struct v4l2_subdev *sd)
{
	v4l2_info(sd, "Nothing to report\n");
	return 0;
}

/* --------------------------------------------------------------------------*/

static const struct v4l2_subdev_core_ops uda1342_core_ops = {
	.log_status = uda1342_log_status,
};

static const struct v4l2_subdev_audio_ops uda1342_audio_ops = {
	.s_routing = uda1342_s_audio_routing,
};

static const struct v4l2_subdev_ops uda1342_ops = {
	.core = &uda1342_core_ops,
	.audio = &uda1342_audio_ops,
};

/* --------------------------------------------------------------------------*/

static int uda1342_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	struct v4l2_subdev *sd;

	/* Check if the adapter supports the needed features */
	if (!i2c_check_functionality(client->adapter, I2C_FUNC_SMBUS_WORD_DATA))
		return -EIO;

	v4l2_info(client, "initializing UDA1342 at address 0x%x on %s\n",
		client->addr, client->adapter->name);

	sd = kmalloc(sizeof(struct v4l2_subdev), GFP_KERNEL);
	if (sd == NULL)
		return -ENOMEM;
	v4l2_i2c_subdev_init(sd, client, &uda1342_ops);

	/* Initialize UDA1342 */
	write_reg(sd, 0x00, 0x8000); /* reset registers */
	write_reg(sd, 0x00, 0x1241); /* select input 1 */

	return 0;
}

static int uda1342_remove(struct i2c_client *client)
{
	struct v4l2_subdev *sd = i2c_get_clientdata(client);

	v4l2_device_unregister_subdev(sd);
	kfree(sd);
	return 0;
}

/* ----------------------------------------------------------------------- */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
static const struct i2c_device_id uda1342_id[] = {
	{ "uda1342", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, uda1342_id);
#endif

static struct v4l2_i2c_driver_data v4l2_i2c_data = {
	.name = "uda1342",
	.probe = uda1342_probe,
	.remove = uda1342_remove,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
	.id_table = uda1342_id,
#endif
};
